/***************************************************************
 * board.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "gggo.h"

void gg_do_reset_mark(void)
  { for(int i = GG_BOARDMIN; i < GG_BOARDMAX; i++) gg_board[i].mark = 0; }

void gg_do_reset_renmark(void)
#if 0
  { int s; for(s = gg_ren[0].next; s != 0; s = gg_ren[s].next) gg_ren[s].mark = 0; }
#else
  { for(int i = 1; i < GG_MAX_REN; i += 1) gg_ren[i].mark = 0; }
#endif

#if USE_HASH
extern bool	gg_is_duplicated(ggMove p, ggColor c);
#endif

extern inline bool gg_is_illegal_ko(ggPoint p, ggColor c);
extern inline bool
gg_is_illegal_ko(ggPoint p, ggColor c) {
  return (c == gg_board_ko_info.color && p == gg_board_ko_info.point)
#if USE_HASH
    || gg_is_duplicated(p, c)
#endif
    ;
}

#define GG_CAN_CAPTURE(p, c) \
  (gg_is_opponent(p, c) && gg_get_liberties(gg_get_ren_id(p)) == 1)

#define GG_NOT_DAMEZUMARI(p, c) \
  (gg_is_friend(p, c) && gg_get_liberties(gg_get_ren_id(p)) != 1)

extern inline bool gg_is_suicide(ggPoint p, ggColor c);
extern inline bool 
gg_is_suicide(ggPoint p, ggColor c) {
  return !gg_is_empty(GG_NORTH(p)) && !gg_is_empty(GG_EAST(p))
    && !gg_is_empty(GG_SOUTH(p)) && !gg_is_empty(GG_WEST(p))
    && !GG_CAN_CAPTURE(GG_NORTH(p), c) && !GG_CAN_CAPTURE(GG_EAST(p), c)
    && !GG_CAN_CAPTURE(GG_SOUTH(p), c) && !GG_CAN_CAPTURE(GG_WEST(p), c)
    && !GG_NOT_DAMEZUMARI(GG_NORTH(p), c) && !GG_NOT_DAMEZUMARI(GG_EAST(p), c)
    && !GG_NOT_DAMEZUMARI(GG_SOUTH(p), c) && !GG_NOT_DAMEZUMARI(GG_WEST(p), c);
}

/*
 * is_illegal(p, color) determines whether the move of <color> at <p> is legal.
 * Returns illegal reason. This can be used as a predicate also.
 */
ggIllegalReason 
gg_is_illegal(ggPoint p, ggColor color) {
  if(gg_is_pass(p))				return GG_IS_LEGAL;
  if(p == GG_RESIGN_MOVE)			return GG_IS_LEGAL;
  assert(gg_is_not_ob(p));
  if(!gg_is_empty(p))				return GG_IS_NOT_EMPTY;
  if(gg_is_illegal_ko(p, color))		return GG_IS_ILLEGAL_KO;
#if 0
  if(!gg_allow_suicide && gg_is_suicide(p, color))	return GG_IS_SUICIDE;
#else
  if(gg_is_suicide(p, color))			return GG_IS_SUICIDE;
#endif
  return GG_IS_LEGAL;
}

extern void	gg_init_stack(void);
extern void	gg_init_timers(void);
#if USE_HASH
extern void	gg_init_hash(void);
extern void	gg_init_trans(void);
#endif

/****************************************
 * 	Basic functions for ren		*
 ****************************************/
void
gg_init_ren(void) {
if(0) gprintf("--> gg_init_ren\n");
  /* Setup free_list for ren. */
  gg_ren_free = GG_REN_ROOT;	/* Root of free ren list. */
  gg_max_ren_free = gg_ren_free;
  gg_ren[0].next = 0;	/* Root of active ren list. */
  for(ggRenID s = GG_REN_ROOT; s < GG_MAX_REN; s++) {
    gg_ren[s].next = s + 1;
    gg_ren[s].mark = 0;
#if SAFETY		/* Not necessary but safety. */
    gg_ren[s].member = 0;
    gg_ren[s].size = 0;
    gg_ren[s].liberties = 0;
    gg_ren[s].color = GG_EMPTY;
    for(int i = 0; i < GG_MAX_LIBS_REN; i += 1) {
      gg_ren[s].liberty[i] = 0;
    }
#endif
  }
  gg_ren[GG_MAX_REN - 1].next = 0;
#if USE_ADJACENT_REN
  for(int i = 0; i < GG_MAX_REN; i += 1) {
    for(int j = 0; j < GG_MAX_REN; j += 1) {
      gg_adjacent_ren[i][j] = false;
    }
  }
#endif
if(0) gprintf("<-- gg_init_ren\n");
  return;
}

void
gg_clear_board(void) {
  ggPoint p;

if(0) gprintf("--> %s()\n", __func__);
  gg_init_stack();
  for(p = 0; p < GG_BOARDSIZE; p += 1) {
    gg_board[p].color = 
	(0 <= GG_X(p) && GG_X(p) < gg_board_size &&
	 0 <= GG_Y(p) && GG_Y(p) < gg_board_size) ? GG_EMPTY : GG_EXTER;
    gg_board[p].mark = 0;
    gg_board[p].ren.id = 0;
    gg_board[p].ren.next = 0;
  }
  gg_init_ren();
#if USE_EMPTY_LIST
#if USE_SPIRAL
{ int x; ggPoint q;
  x = 0;
  p = 0;
  q = GG_TENGEN;
  for(int i = 0; i < gg_board_size * gg_board_size; i += 1) {
    for(int j = 0; j < 1 + i / 2; j += 1) {
      if(gg_is_empty(q)) {
if(0) gprintf("%1m ", q);
	gg_empty_list[p] = q;
	p = q;
      }
      q += gg_direction[x];
      if(gg_is_exter(q))	goto next;
    }
    x += 2; if(x >= 8)	x = 0;
  }
}
next:
if(0) gputc('\n');
#else
  p = 0;
  for(ggPoint q = GG_BOARDMIN; q < GG_BOARDMAX; q += 1) {
    if(!gg_is_empty(q))	continue;
    gg_empty_list[p] = q;
    p = q;
  }
#endif
  gg_empty_list[p] = 0;
#endif
  gg_bw_captured[GG_C2X(GG_WHITE)] = 0;
  gg_bw_captured[GG_C2X(GG_BLACK)] = 0;
  gg_board_ko_info.point = GG_NO_MOVE;
  gg_board_ko_info.color = GG_EMPTY;
  gg_previous_move = GG_NO_MOVE;
  gg_previous_move2 = GG_NO_MOVE;
  gg_movenum = 0;
  gg_n_pass = 0;
  gg_gameend = false;
#if USE_HASH
  gg_init_hash();
  gg_init_trans();
#endif
  gg_init_timers();
if(0) gprintf("<-- %s\n", __func__);
  return;
}

/* Bit map table is enough.
 * Caution: this function returns incorrect value sometime.
 * Use this in limited conditions.
 */
bool gg_is_rappa(ggPoint p);
bool
gg_is_rappa(ggPoint p) {

  /* Pack 4 adjacent intersections into 8 bits and jump! */
  switch(((((((gg_get_color(p + GG_N_NORTH)) << 2) |
	       gg_get_color(p + GG_N_EAST)) << 2) |
	       gg_get_color(p + GG_N_SOUTH)) << 2) |
	       gg_get_color(p + GG_N_WEST)) {
    /* 1 liberty */
    case 0x15: case 0x45: case 0x51: case 0x54: /* 3 Ws and 1 liberty */
    case 0x2a: case 0x8a: case 0xa2: case 0xa8:	/* 3 Bs and 1 liberty */
    case 0x35: case 0x1d: case 0x17:	/* 2 Ws, 1 X and 1 lib */
    case 0xc5: case 0x4d: case 0x47:
    case 0xd1: case 0x71: case 0x53:
    case 0xd4: case 0x74: case 0x5c:
    case 0x3a: case 0x2e: case 0x2b:	/* 2 Bs, 1 X and 1 lib */
    case 0xca: case 0x8e: case 0x8b:
    case 0xe2: case 0xb2: case 0xa3:
    case 0xe8: case 0xb8: case 0xac:
    case 0x1f: case 0x4f:		/* 1 White, 2 Xs and 1 lib */
    case 0x3d: case 0x7c:
    case 0xf1: case 0xf4:
    case 0xc7: case 0xd3:
    case 0x2f: case 0x8f:		/* 1 Black, 2 Xs and 1 lib */
    case 0x3e: case 0xbc:
    case 0xf2: case 0xf8:
    case 0xcb: case 0xe3:
      return true;
    break;
    /* No liberty */
    case 0x55: case 0xaa:			/* 4 Ws, 4 Bs */
    case 0xd5: case 0x75: case 0x5d: case 0x57:	/* 3 Ws and 1 X */
    case 0xea: case 0xba: case 0xab: case 0xae:	/* 3 Bs and 1 X */
     /* FIXME */
      return true;
    break;
  
    default:
      return false;
    break;

  /* Impossible cases (31)
   */
    case 0xff:					/* 4 exteriors */
    case 0x3f: case 0xfc: case 0xf3: case 0xcf:	/* 3 exteriors */
    case 0x7f: case 0xfd: case 0xf7: case 0xdf:	/* + 1 white */
    case 0xbf: case 0xfe: case 0xfb: case 0xef:	/* + 1 black */
    case 0x33: case 0xcc:			/* 2 Xs in face to face */
    case 0x37: case 0x73: case 0xcd: case 0xdc:	/* + 1 white */
    case 0x3b: case 0xb3: case 0xce: case 0xec:	/* + 1 black */
    case 0x77: case 0xdd:			/* + 2 whites */
    case 0xbb: case 0xee:			/* + 2 blacks */
    case 0x7b: case 0xb7: case 0xed: case 0xde:	/* + white and black */
      assert(false);
    break;
  }
  return false;
}

void	gg_show_ren_info(const ggPoint p);
void
gg_show_ren_info(const ggPoint p) {
  ggRenID s;

  assert(gg_is_not_ob(p));
  s = gg_get_ren_id(p);
  assert(s != 0);
  printf("ID: %d\nColor: %C\n", s, gg_color_ren(s));
  printf("Size: %d\nLiberties: %d\n", gg_size_ren(s), gg_libs_ren(s));
  printf("Members:");
  for(int k = 1, p = gg_member_ren(s); p != 0; p = gg_next_member_ren(p)) {
    if(++k % 5 == 0) putchar('\n');
    mprintf(" %1m", p);
  }
  putchar('\n');
  putchar('\n');
  return;
}

static uint16_t	gg_captured_stones;
static ggPoint	gg_captured_pos;
static ggRenID	*gg_dirty_ren0, gg_dirty_ren[GG_MAX_REN + 1];

#define MAX_LIBS_REN	GG_MAX_LIBS_REN

#define Color		ggColor
#define Move		ggMove
#define RenID		ggRenID
#define Size		ggSize

#define update_ren	gg_update_ren
#define create_ren	gg_create_ren
#define last_stone	gg_last_stone
#define last_ren	gg_last_ren
#define collect_ren	gg_collect_ren
#define remove_ren	gg_remove_ren
#define join_ren	gg_join_ren
#define find_libs_ren	gg_find_libs_ren
#define add_stone	gg_add_stone

#define position	gg_position
#define ren_free	gg_ren_free
#define max_ren_free	gg_max_ren_free
#define dirty_ren	gg_dirty_ren
#define dirty_ren0	gg_dirty_ren0
#define captured_stones	gg_captured_stones
#define captured_pos	gg_captured_pos

#define get_color	gg_get_color
#define set_color	gg_set_color
#define clear_color	gg_clear_color
#define is_empty	gg_is_empty
#define is_onboard	gg_is_onboard
#define is_onboard2	gg_is_onboard2
#define is_not_ob	gg_is_not_ob
#define is_b_or_w	gg_is_b_or_w
#define is_opponent	gg_is_opponent
#define is_friend	gg_is_friend
#define opposit_color	gg_opposit_color
#define x2c		gg_x2c

#define id_ren		gg_id_ren
#define next_member_ren	gg_next_member_ren
#define next_ren	gg_next_ren
#define color_ren	gg_color_ren
#define size_ren	gg_size_ren
#define libs_ren	gg_libs_ren
#define member_ren	gg_member_ren
#define liberty_ren	gg_liberty_ren

#define mark		gg_mark
#define is_marked	gg_is_marked
#define reset_mark	gg_reset_mark
#define mark_ren	gg_mark_ren
#define is_marked_ren	gg_is_marked_ren
#define reset_renmark	gg_reset_renmark

#define EMPTY		GG_EMPTY
#define MIN_COLOR	GG_MIN_COLOR
#define NORTH		GG_NORTH
#define EAST		GG_EAST
#define SOUTH		GG_SOUTH
#define WEST		GG_WEST
#define FOR_ALL_POINT_IN_REN		GG_FOR_ALL_POINT_IN_REN
#define FOR_4_ADJACENT_POINT		GG_FOR_4_ADJACENT_POINT

#define FATAL		GG_FATAL
#define BLACK		GG_BLACK
#define WHITE		GG_WHITE
#define EXTER		GG_EXTER
#define PASS_MOVE	GG_PASS_MOVE
#define RESIGN_MOVE	GG_RESIGN_MOVE
#define POS		GG_POS
#define X		GG_X
#define Y		GG_Y

#if USE_EMPTY_LIST
#define empty_list	gg_empty_list
#define remove_from_empty_list	gg_remove_from_empty_list
#endif

#if USE_ADJACENT_REN
#define adjacent_ren	gg_adjacent_ren
#endif

#include "update_ren.c"

void
gg_play_and_analyze(ggMove move, ggColor color) {
  ggRenID self;

if(0) gprintf("--> gg_play_and_analyze(%1m,%C)\n", move, color);

  assert(gg_is_not_ob(move));

  /* Update board and rens with gathering some info.
   */
  gg_captured_pos = GG_NO_MOVE;
  self = gg_update_ren(move, color);

  /* Check ko move.
   */
  if(gg_captured_stones == 1 && gg_ren[self].liberties == 1 &&
	    gg_ren[self].size == 1) {
    assert(gg_captured_pos != GG_NO_MOVE);
    gg_board_ko_info.point = gg_captured_pos;
    gg_board_ko_info.color = gg_opposit_color(color);
if(0) gprintf("--- gg_play_and_analyze detect KO %1m %C\n",
		  gg_board_ko_info.point, gg_board_ko_info.color);
  } else {
    gg_board_ko_info.color = GG_EMPTY;
  }
  /* Update captured stones info. */
  gg_bw_captured[GG_C2X(color)] += gg_captured_stones;

if(0) gprintf("<-- gg_play_and_analyze\n");
  return;
}
