/***************************************************************
 * cs.h
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef _CS_H_
#define _CS_H_

#include <signal.h>
#include <time.h>
#include "config.h"
#include "mc.h"
#include "uct.h"

#if USE_SPU
#include <libspe2.h>
#endif

#ifndef __SPU__
#include <pthread.h>
#include "queue.h"
#endif

#define CS_PACKET_VERSION	((uint8_t)1)
#define CS_PACKET_POSITION	((uint8_t)1)
#define CS_PACKET_RESULT	((uint8_t)2)

/****************************************
 *	Client-Server Parallel MC/UCT	*
 *	over Network using UDP/IP	*
 ****************************************/
#if 0
typedef enum {
  CS_CLIENT_SHARE = 1,
  CS_CLIENT_LAN,
  CS_CLIENT_SPU,
} ggCliendType;			/* Not used. */
#endif

typedef enum {
  CS_IN_USE = 1,
} csPacketStatus;

typedef enum {
  CS_READY = 1,
  CS_HALT,
  CS_EXEC,
  CS_SCORE,
  CS_SIZE,
  CS_DONE,
  CS_PROFILE,
} ggCommand;			/* Also used for replies. */

/* Packed mcRave */
#if USE_RAVE
typedef struct {		/* Caution! just an image */
  int16_t	n_rave;
  mcMove	move[0];		/* move[n_rave] */
  mcColor	color[0];		/* color[n_rave] */
  struct {
    mcMove	m;
    mcColor	c;
  } space[MC_MAX_MOVENUM];		/* Real space */
} csRave		__attribute__((aligned(16)));
#endif

typedef struct {
  uint16_t	size;		/* Size of packet */
  uint8_t	version, type;
  uint32_t	t_send, t_recv, t_done;
  uint32_t	generation;
  uint32_t	playout;	/* Optional */
  uint32_t	node;
  ggArc		optarc;
  ggScore	score;
#if USE_RAVE
  mcRave	rave;
#endif
} mcResult;

/************************
 *   Packed packet	*
 ************************/
typedef struct {
  mcMove	first_move;
  uint8_t	board_size;
  csPackedPosition	position;
} csPosition;

typedef struct {
  ggScore	score;
#if USE_RAVE
  csRave	rave;
#endif
} csResult;

typedef struct {
  uint16_t	size;		/* Size of packet */
  uint8_t	version, type;
  uint32_t	t_send, t_recv, t_done;
  uint32_t	generation;
  uint32_t	playout;	/* Optional */
  uint32_t	node;
  ggArc		optarc;
  union {
    csPosition	position;
    csResult	result;
  } u;
} csPacket;

#ifndef __SPU__
typedef struct {
  mcPosition	position;
  uint32_t	node;
  uint32_t	playout;
  ggArc		optarc;
  ggMove	first_move;
#if USE_RAVE
  mcRave	rave;
#endif
#if USE_SPU
  spe_context_ptr_t	spe;
#else
  pthread_t	thread;
#endif
  ggQueue	ready_queue;
  ggQueue	score_queue;
  ggQueue	command;
  ggRandomSeed	seed;
  uint16_t	id;
} csThread;
#endif

typedef struct {
  ggRandomSeed	seed;
#if 1
  uint8_t	id;
  int8_t	packet;
#else
  unsigned int	id;
  int		packet;
  int		dummy[100];
#endif
#ifndef __SPU__
  pthread_t	thread;
#if USE_SPU
  spe_context_ptr_t	spe;
#else
  ggQueue	ready_queue;
  ggQueue	score_queue;
  ggQueue	command;
#endif
#endif
} csSimulator;

/**************************************/
#define cs_id	cs_index
static inline int	cs_command(ggMessage msg);
extern inline int	cs_index(ggMessage msg);
extern inline int	cs_value(ggMessage msg);
static inline ggMessage	cs_message(int value, int id, ggCommand command);
static inline int	cs_command(ggMessage msg) { return msg & 0xff; }
extern inline int	cs_index(ggMessage msg) { return (msg >> 8) & 0xff; }
extern inline int	cs_value(ggMessage msg) { return (int32_t)msg >> 16; }
static inline ggMessage	cs_message(int value, int id, ggCommand command) {
  return (value << 8 | id) << 8 | command; }

extern void	cs_exit_perror(const char *msg);
extern void	cs_usleep(int usec);
extern void	cs_nsleep(long nsec);
extern uint32_t	cs_timestamp(void);
extern int32_t	cs_timesub(uint32_t x, uint32_t y);

#ifndef __SPU__
extern void	mc_do_simulation(csSimulator *self);
extern volatile bool	cs_timeup;
extern void	cs_create_timer(struct sigevent *evp, timer_t *id,
		    struct sigaction *act, void (*handler)(int), int signum);
extern void	cs_delete_timer(timer_t id, struct sigaction *act);
extern void	cs_set_timer(ggTime t, struct itimerspec *value, timer_t id);
extern void	cs_init_client(void);
extern uint32_t	cs_do_playout(ggUctNode *root, uint32_t n_playout);
#endif

#endif	/* _CS_H_ */
