/***************************************************************
 * gggo.h
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef _GGGO_H_
#define _GGGO_H_

#include <assert.h>
#include "config.h"
#include "types.h"

extern inline int	GG_X(ggPoint p);
extern inline int	GG_Y(ggPoint p);
extern inline int	GG_X(ggPoint p) { return (p) % GG_N - 1; }
extern inline int	GG_Y(ggPoint p) { return (p) / GG_N - 1; }

#define GG_NORTH(p)	((p) + GG_N_NORTH)
#define GG_SOUTH(p)	((p) + GG_N_SOUTH)
#define GG_EAST(p)	((p) + GG_N_EAST)
#define GG_WEST(p)	((p) + GG_N_WEST)
#define GG_NE(p)	((p) + GG_N_NORTH + GG_N_EAST)
#define GG_NW(p)	((p) + GG_N_NORTH + GG_N_WEST)
#define GG_SE(p)	((p) + GG_N_SOUTH + GG_N_EAST)
#define GG_SW(p)	((p) + GG_N_SOUTH + GG_N_WEST)
#define GG_NN(p)	((p) + GG_N_NORTH + GG_N_NORTH)
#define GG_EE(p)	((p) + GG_N_EAST + GG_N_EAST)
#define GG_SS(p)	((p) + GG_N_SOUTH + GG_N_SOUTH)
#define GG_WW(p)	((p) + GG_N_WEST + GG_N_WEST)

/****************************************
 *		Interface		*
 ****************************************/
extern int	gg_board_size;
extern float	gg_komi;
extern int	gg_allow_suicide;
extern int	gg_resign_allowed;
extern int	gg_never_resign;
extern int	gg_never_pass;
#if USE_LADDER_MODE
extern int	gg_use_ladder_mode;
#endif

extern uint16_t	gg_bw_captured[GG_N_COLOR];
extern uint16_t	gg_movenum;

extern uint8_t	gg_n_pass;
extern ggMove	gg_previous_move;
extern ggMove	gg_previous_move2;
extern ggColor	gg_previous_color;
extern ggBool	gg_gameend;
extern ggKoInfo	gg_board_ko_info;
extern ggBoard	gg_board[GG_BOARDSIZE];
extern ggRen	gg_ren[GG_MAX_REN];
extern ggRenID	gg_ren_free;
extern ggRenID	gg_max_ren_free;
#if USE_EMPTY_LIST
extern ggPoint	gg_empty_list[GG_BOARDSIZE];
#endif

extern ggPosition *gg_position_sp;

extern const ggDirection	gg_direction[8];
extern const ggDirection	g2_direction[2][8];

#if USE_HASH
extern ggHash	gg_hash, gg_zobrist[][GG_N_COLOR];
#endif
extern void	gg_play_move(ggMove move, ggColor color);
extern void	gg_play_and_analyze(ggMove move, ggColor color);
extern bool	gg_undo_move(int n);
extern ggMove	gg_genmove(ggColor color);
extern ggScoreValue	gg_finish_and_score(ggRandomSeed seed);
extern ggIllegalReason	gg_is_illegal(ggMove p, ggColor c);
extern bool	gg_is_self_atari(ggMove m, ggColor color);
extern bool	gg_is_rappa(ggMove p);
extern void	gg_clear_board(void);
extern void	gg_init_ren(void);
extern ggRenID	gg_update_ren(ggMove p, ggColor c);

extern void	gg_init_gen_rand(uint32_t seed);

/*	Basic functions for board and ren
 */
#define gg_color(p)		gg_board[(p)].color
#define gg_id_ren(p)		gg_board[(p)].ren.id
#define gg_next_member_ren(p)	gg_board[(p)].ren.next
#define gg_color_ren(s)		gg_ren[(s)].color
#define gg_size_ren(s)		gg_ren[(s)].size
#define gg_next_ren(s)		gg_ren[(s)].next
#define gg_member_ren(s)	gg_ren[(s)].member
#define gg_libs_ren(s)		gg_ren[(s)].liberties
#define gg_liberty_ren(s)	gg_ren[(s)].liberty

extern inline ggColor gg_get_color(ggPoint p);
extern inline ggColor gg_set_color(ggPoint p, ggColor c);
extern inline ggColor gg_clear_color(ggPoint p);
extern inline ggRenID gg_get_ren_id(ggPoint p);
extern inline bool gg_is_not_ob(ggPoint p);
extern inline bool gg_is_exter(ggPoint p);
extern inline bool gg_is_onboard(ggPoint p);
extern inline bool gg_is_onboard2(ggPoint p);
extern inline bool gg_is_empty(ggPoint p);
extern inline bool gg_is_b_or_w(ggPoint p);
extern inline bool gg_is_friend(ggPoint p, ggColor c);
extern inline ggColor gg_opposit_color(ggColor c);
extern inline bool gg_is_opponent(ggPoint p, ggColor c);
extern inline ggUint gg_get_liberties(ggRenID id);
extern inline int gg_count_liberty(ggPoint p);

extern inline ggColor gg_get_color(ggPoint p)
    { return gg_board[p].color; }
extern inline ggColor gg_set_color(ggPoint p, ggColor c) {
#if USE_HASH
  gg_hash ^= gg_zobrist[p][GG_C2X(c)];
#endif
  return gg_board[p].color = c; }
extern inline ggColor gg_clear_color(ggPoint p) {
#if USE_HASH
  gg_hash ^= gg_zobrist[p][GG_C2X(gg_board[p].color)];
#endif
  return gg_board[p].color = GG_EMPTY; }
extern inline ggRenID gg_get_ren_id(ggPoint p)
    { return gg_board[p].ren.id; }

extern inline bool gg_is_not_ob(ggPoint p)
    { return p < GG_BOARDSIZE; }
extern inline bool gg_is_exter(ggPoint p)
    { return gg_board[p].color == GG_EXTER; }
extern inline bool gg_is_onboard(ggPoint p)
    { return gg_board[p].color != GG_EXTER; }
extern inline bool gg_is_onboard2(ggPoint p) {
  return 0 <= GG_X(p) && GG_X(p) < gg_board_size &&
	 0 <= GG_Y(p) && GG_Y(p) < gg_board_size; }
extern inline bool gg_is_empty(ggPoint p)
    { return gg_board[p].color == GG_EMPTY; }
extern inline bool gg_is_b_or_w(ggPoint p)
    { ggColor c = gg_board[p].color; return c == GG_BLACK || c == GG_WHITE; }
extern inline bool gg_is_friend(ggPoint p, ggColor c)
    { return gg_board[p].color == c; }
extern inline ggColor gg_opposit_color(ggColor c)
    { return c ^ (GG_WHITE ^ GG_BLACK); }
extern inline bool gg_is_opponent(ggPoint p, ggColor c)
    { return gg_board[p].color == gg_opposit_color(c); }

extern inline ggUint gg_get_liberties(ggRenID id)
    { return gg_ren[id].liberties; }

/* TODO Use pre-compiled table with 4x4x4x4 entries. */
extern inline int
gg_count_liberty(ggPoint p) {
  int n;
  assert(gg_is_onboard(p));
  n = 0;
#if 0
  for(int x = 0; x < 8; x += 2) {
    ggPoint q = p + gg_direction[x];
    if(gg_is_empty(q))	n += 1;
  }
#else
#define GG_LIBERTIES(p) (gg_is_empty(p) ? 1 : 0)
  n += GG_LIBERTIES(GG_NORTH(p));
  n += GG_LIBERTIES(GG_EAST(p));
  n += GG_LIBERTIES(GG_SOUTH(p));
  n += GG_LIBERTIES(GG_WEST(p));
#endif
  return n;
}

extern inline void gg_mark(ggPoint p);
extern inline void gg_clear_mark(ggPoint p);
extern inline bool gg_is_marked(ggPoint p);
extern inline void gg_reset_mark(void);
extern inline void gg_mark_ren(ggRenID s);
extern inline bool gg_is_marked_ren(ggRenID s);
extern inline bool gg_is_marked_ren(ggRenID s);
extern inline void gg_reset_renmark(void);
extern inline bool gg_is_pass(ggMove m);

/*	Marking functions for board
 */
extern ggMark	gg_mark_num;		/* private variable */
extern inline void gg_mark(ggPoint p)
    { gg_board[p].mark = gg_mark_num; }
extern inline void gg_clear_mark(ggPoint p)
    { gg_board[p].mark = 0; }
extern inline bool gg_is_marked(ggPoint p)
    { return gg_board[p].mark == gg_mark_num; }
extern void gg_do_reset_mark(void);		/* private function */
extern inline void gg_reset_mark(void)
    { while(++gg_mark_num == 0) gg_do_reset_mark(); }

extern ggMark	gg_mark_ren_num;		/* private variable */
extern inline void gg_mark_ren(ggRenID s)
    { gg_ren[s].mark = gg_mark_ren_num; }
extern inline bool gg_is_marked_ren(ggRenID s)
    { return gg_ren[s].mark == gg_mark_ren_num; }
extern void gg_do_reset_renmark(void);		/* private function */
extern inline void gg_reset_renmark(void)
    { while(++gg_mark_ren_num == 0) gg_do_reset_renmark(); }

extern inline bool gg_is_pass(ggMove m) { return m == GG_PASS_MOVE; }

/*	Control Structure Macros
 */
#define GG_FOR_ALL_ADJACENT(q, p, x) \
  for((x) = 0; (q) = (p) + gg_direction[(x)], (x) < 8; (x) += 2)

#define GG_FOR_4_ADJACENT_POINT(q, p) \
  for(int z = 0; (q) = (p) + gg_direction[z], z < 8; z += 2)

#define GG_FOR_8_ADJACENT_POINT(q, p) \
  for(int z = 0; (q) = (p) + gg_direction[z], z < 8; z += 1)

#define GG_FOR_ALL_POINT(p) \
  for((p) = GG_BOARDMIN; (p) < GG_BOARDMAX; p += 1)

#define GG_FOR_ALL_COLOR(c) \
  for((c) = GG_MIN_COLOR; (c) <= GG_MAX_COLOR;	\
      (c) += GG_MAX_COLOR - GG_MIN_COLOR)

#define GG_FOR_ALL_REN(s) \
  for((s) = gg_ren[0].next; (s) != 0; (s) = gg_ren[(s)].next)

#define GG_FOR_ALL_POINT_IN_REN(p, s) \
  for((p) = gg_ren[(s)].member; (p) != 0; (p) = gg_board[(p)].ren.next)

#if USE_EMPTY_LIST
#define GG_FOR_ALL_EMPTY_POINT(p)	\
  for((p) = gg_empty_list[0]; (p) != 0; (p) = gg_empty_list[p])
#else
#define GG_FOR_ALL_EMPTY_POINT(p)	\
  for((p) = GG_BOARDMIN; (p) < GG_BOARDMAX; p += 1)
#endif

#if USE_ADJACENT_REN
#define	gg_adjacent_ren	(gg_position_sp->adjacent_ren)
#define GG_BEGIN_FOR_ALL_ADJACENT_REN(t, s)	\
  for((t) = 0; (t) < max_ren_free; (t) += 1) {	\
    if(adjacent_ren[s][t] == 0)	continue;
#define GG_END_FOR_ALL_ADJACENT_REN	}
#endif

/************************************************
 *	For GTP command and print utils		*
 ************************************************/
extern inline char gg_x2c(int i);
extern inline char
gg_x2c(int i) { return 'A' + i + ('A' + i >= 'I' ? 'J' - 'I' : 0); }

#if USE_ORACLE
#define GG_ORACLE_LINESIZE	256
extern void	gg_put_oracle(const char *format, ...);
extern char	*gg_get_oracle(char *line);
extern FILE	*gg_oracle_from, *gg_oracle_to;
extern char	gg_oracle[GG_ORACLE_LINESIZE];
#endif

#endif /* _GGGO_H_ */
