/***************************************************************
 * pack.h
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef _PACK_H_
#define _PACK_H_

#include "cs.h"
#include "mc.h"
#ifdef __SPU__			/* CELL's bytesex is big endian. */
#define ntohl(x)	(x)
#define ntohs(x)	(x)
#define htonl(x)	(x)
#define htons(x)	(x)
#else
#include <arpa/inet.h>
#endif

#define CS_PACK(x)	(sizeof(x) == 1 ? (uint8_t)(x) : \
    sizeof(x) == 2 ? (uint16_t)htons(x) : (uint32_t)htonl(x))
#define CS_UNPACK(x) (sizeof(x) == 1 ? (uint8_t)(x) : \
    sizeof(x) == 2 ? (uint16_t)ntohs(x) : (uint32_t)ntohl(x))

extern uintptr_t
	cs_pack_position(csPackedPosition *dst, const mcPosition *src);
extern void
	cs_unpack_position(mcPosition *dst, const csPackedPosition *src);
#if USE_RAVE
extern uintptr_t cs_pack_rave(csRave *dst, const mcRave *src);
extern void	cs_unpack_rave(mcRave *dst, const csRave *src);
#endif

extern void	dump_mc_position(mcPosition *p, mcMove first_move, FILE *s);
#if USE_RAVE
extern void	dump_mc_rave(mcRave *p, FILE *s);
extern void	dump_cs_rave(csRave *p, FILE *s);
#endif

#endif	/* _PACK_H_ */
