/***************************************************************
 * printutils.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdio.h>

#if ! HAS_FLOCK
#define putc_unlocked	putc
#define fputs_unlocked	fputs
#endif

static char
color2char(int color) {
  char c;
  if(color == BLACK)		c = 'B';
  else if(color == WHITE)	c = 'W';
  else if(color == EMPTY)	c = '.';
  else if(color == EXTER)	c = 'X';
  else				c = '?';
  return c;
}

void 
vgprintf(FILE *stream, const char *format, va_list ap) {
  char *modifier0, modifier[64];	/* FIXME Dirty hack. */

  while(*format != '\0') {
    if(*format++ == '%') {
      modifier0 = &modifier[0];
      *modifier0++ = '%';
      while(isdigit((unsigned char)*format) || *format == '+' ||
	*format == '-' || *format == '.' || *format == '#' || *format == ' ' ||
	*format == '\'' || *format == 'I' || *format == 'l') {
	assert(modifier0 < &modifier[64 - 2]);
	*modifier0++ = *format++;
      }
      *modifier0++ = *format;
      *modifier0 = '\0';
      switch (*format++) {
      case 'c':
      case 'd': case 'u': case 'i':
      case 'o': case 'x': case 'X':
      	fprintf(stream, modifier, va_arg(ap, int));
	break;
      case 'f': case 'F': case 'e': case 'E':
      case 'g': case 'G': case 'a': case 'A':
      	fprintf(stream, modifier, va_arg(ap, double));
	break;
      case 's':		/* Maybe same as 'p' */
      	fprintf(stream, modifier, va_arg(ap, char *));
	break;
      case 'p':
      	fprintf(stream, modifier, va_arg(ap, void *));
	break;
      case 'C':
	fprintf(stream, "%c", color2char(va_arg(ap, int)));
	break;
      case 'b':
	fputs_unlocked((va_arg(ap, int) == 0) ?  "false" : "true", stream);
	break;
      case 'B':
	fputs_unlocked((va_arg(ap, int) == 0) ?  "FALSE" : "TRUE", stream);
	break;
      case 'm': case 'M':
	if(*(format - 2) == '2' || *(format - 2) == '%') {
	  int m = va_arg(ap, int);
	  int n = va_arg(ap, int);
	  int pos = POS(m, n);
	  if(pos == PASS_MOVE) {
	    fputs_unlocked("pass", stream);
	  } else if(pos == RESIGN_MOVE) {
	    fputs_unlocked("resign", stream);
	  } else if(is_not_ob(pos) && is_onboard2(pos)) {
	    putc_unlocked(x2c(m), stream);
	    fprintf(stream, (*(format - 1) == 'm' ? "%d" : "%-2d"), n + 1);
	  } else {
	    fprintf(stream, "[%d,%d]", m, n);
	  }
	} else if(*(format - 2) == '1') {	/* "1m" or "1M" */
	  int pos = va_arg(ap, int);
	  int m = X(pos);
	  int n = Y(pos);
	  if(pos == PASS_MOVE) {
	    fputs_unlocked("pass", stream);
	  } else if(pos == RESIGN_MOVE) {
	    fputs_unlocked("resign", stream);
	  } else if(is_not_ob(pos) && is_onboard2(pos)) {
	    putc_unlocked(x2c(m), stream);
	    fprintf(stream, "%d", n + 1);
	  } else {
	    fprintf(stream, "[%d]", pos);
	  }
	} else {
	  fprintf(stream, "\nUnknown format string \"%s\"\n", &modifier[0]);
	}
	break;
      default:
	putc_unlocked(*(format - 1), stream);
	break;
      }
    } else {
      putc_unlocked(*(format - 1), stream);
    }
  }
  return;
}

int 
gprintf(const char *format, ...) {
  va_list ap;
#if HAS_FLOCK
  flockfile(stderr);
#endif
  va_start(ap, format);
  vgprintf(stderr, format, ap);
  va_end(ap);
#if HAS_FLOCK
  funlockfile(stderr);
#endif
  return 1;
}

int
mprintf(const char *format, ...) {
  va_list ap;
#if HAS_FLOCK
  flockfile(stdout);
#endif
  va_start(ap, format);
  vgprintf(stdout, format, ap);
  va_end(ap);
#if HAS_FLOCK
  funlockfile(stdout);
#endif
  return 1;
}

int 
gfprintf(FILE *stream, const char *format, ...) {
  va_list ap;
#if HAS_FLOCK
  flockfile(stream);
#endif
  va_start(ap, format);
  vgprintf(stream, format, ap);
  va_end(ap);
#if HAS_FLOCK
  funlockfile(stream);
#endif
  return 1;
}
