/***************************************************************
 * queue.h
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef _GG_QUEUE_
#define _GG_QUEUE_

#ifndef __SPU__

#include <pthread.h>
#include "types.h"

#define GG_MAX_QUEUE_LENGTH	4

typedef struct {
  ggMessage		data[GG_MAX_QUEUE_LENGTH];
  pthread_mutex_t	mutex;
  pthread_cond_t	cond_not_empty;
  pthread_cond_t	cond_not_full;
  uint8_t	head, tail, len;
} ggQueue;

extern int	gg_init_queue(ggQueue *queue);
extern int	gg_push(int32_t msg, ggQueue *queue);
extern int32_t	gg_pop(ggQueue *queue, int *rc);
extern int	gg_queue_length(ggQueue *queue, int *rc);
extern int	gg_clear_queue(ggQueue *queue);
extern bool	gg_queue_is_empty(ggQueue *queue, int *rc);
extern bool	gg_queue_is_full(ggQueue *queue, int *rc);

#endif	/* __SPU__ */

#endif
